@@ -64,6 +64,10 @@ class BinRegexps:
 
 def fucking_magic(f):
     # m = magic.detect_from_filename(f)
+    if "ld.so" in f:
+        wtf = 1 
+        pass
+
     if not os.path.exists(f):
         return ''
 
@@ -114,6 +118,10 @@ class FoldersSpec:
     folders:   list
 
 
+@dc.dataclass
+class ModulesSpec:
+    modules:   list
+
 
 
 class TerrariumAssembler:
@@ -152,6 +160,7 @@ class TerrariumAssembler:
         # ap.add_argument('--stage-build-nuitka', default=False, action='store_true', help='Compile Nuitka packages')
         ap.add_argument('--stage-build-and-pack', default='', type=str, help='Install, build and pack')
         ap.add_argument('--stage-download-all', default=False, action='store_true', help='Download all — sources, packages')
+        ap.add_argument('--stage-my-source-changed', default='', type=str, help='Fast rebuild/repack if only pythonsourcechanged')
         ap.add_argument('--stage-all', default='', type=str, help='Install, build and pack')
         ap.add_argument('--stage-pack', default='', type=str, help='Stage pack to given destination directory')
         ap.add_argument('specfile', type=str, help='Specification File')
@@ -168,6 +177,14 @@ class TerrariumAssembler:
             self.args.stage_build_nuitka = True
             self.args.stage_pack = self.args.stage_build_and_pack
 
+        if self.args.stage_my_source_changed:
+            self.args.stage_checkout = True
+            self.args.stage_download_wheels = True
+            self.args.stage_build_wheels = True
+            self.args.stage_install_wheels = True
+            self.args.stage_build_nuitka = True
+            self.args.stage_pack = self.args.stage_my_source_changed
+
         if self.args.stage_download_all:
             self.args.stage_download_rpms = True
             self.args.stage_checkout = True
@@ -200,8 +217,8 @@ class TerrariumAssembler:
         )
 
         self.need_packages = ['patchelf', 'ccache', 'gcc', 'gcc-c++', 'gcc-gfortran', 'chrpath', 
-                                'python3-wheel', 'python3-pip', 'python3-devel', 
-                                'genisoimage', 'makeself']
+                                'python3-wheel', 'python3-pip', 'python3-devel', 'python3-yaml',
+                                'genisoimage', 'makeself', 'dnf-utils']
 
         nflags_ = {}
         if 'nuitka' in spec:
@@ -234,6 +251,19 @@ class TerrariumAssembler:
 
         pass
 
+
+    def packages2list(self, pl):
+        pl_ = []
+        for node in pl:
+            if isinstance(node, str):
+                pl_.append(node)
+            if isinstance(node, dict):
+                if 'name' in node:
+                    pl_.append(node['name'])
+        return pl_            
+
+
+
     def lines2sh(self, name, lines, stage=None):
         import stat
         os.chdir(self.curdir)
@@ -274,13 +304,71 @@ set -x
         #     return
         # out_dir = os.path.join(self.out_dir)
         tmpdir = os.path.join(self.curdir, "tmp/ta")
+        tmpdir_ = os.path.relpath(tmpdir)
         bfiles = []
+
+        #First pass
+        module2build = {}
+        standalone2build = []
+        referenced_modules = set()
+
         for target_ in self.nuitkas.builds:
-            # outputname = target_.utility
-            # if "outputname" in target_:
-            #     outputname = target_.outputname
-            nflags = self.nuitkas.get_flags(tmpdir)
-            target_dir = os.path.join(tmpdir, target_.utility+'.dist')
+            if 'module' in target_:
+                module2build[target_.module] = target_
+            else:
+                standalone2build.append(target_)
+                if 'modules' in target_:
+                    referenced_modules |= set(target_.modules)
+                    for it_ in target_.modules:
+                        if it_ not in module2build:
+                            module2build[it_] = edict({'module':it_})
+
+        #processing modules only 
+
+        for outputname, target_ in module2build.items():
+            block_modules = None
+            if 'block_modules' in target_:
+                block_modules = target_.block_modules
+
+            nflags = self.nuitkas.get_flags(os.path.join(tmpdir, 'modules', outputname), target_)
+            if not nflags:
+                continue
+            target_dir = os.path.join(tmpdir, outputname + '.dist')
+            target_dir_ = os.path.relpath(target_dir, start=self.curdir)
+            target_list = target_dir_.replace('.dist', '.list')
+            tmp_list = '/tmp/module.list'
+            source_dir = dir4mnode(target_)
+            flags_ = ''
+            if 'flags' in target_:
+                flags_ = target_.flags
+            lines = []
+            build_name = 'build_module_' + outputname
+            lines.append("""
+export PATH="/usr/lib64/ccache:$PATH"
+find %(source_dir)s -name "*.py" | xargs -i{}  cksum {} > %(tmp_list)s
+if cmp -s %(tmp_list)s %(target_list)s
+then
+    echo "Module '%(outputname)s' looks unchanged" 
+""" % vars())
+            lines.append(R"""
+else
+    python3 -m nuitka  %(nflags)s %(flags_)s  2>&1 >%(build_name)s.log
+    RESULT=$?
+    if [ $RESULT == 0 ]; then
+        cp %(tmp_list)s %(target_list)s
+    fi
+fi
+""" % vars())
+            self.fs.folders.append(target_dir)
+            self.lines2sh(build_name, lines, None)
+            bfiles.append(build_name)
+
+        for target_ in standalone2build:
+            srcname = target_.utility
+            outputname = target_.utility
+            nflags = self.nuitkas.get_flags(tmpdir, target_)
+            target_dir = os.path.join(tmpdir, outputname + '.dist')
+            target_dir_ = os.path.relpath(target_dir, start=self.curdir)
             src_dir = os.path.relpath(self.src_dir, start=self.curdir)
             src = os.path.join(src_dir, target_.folder, target_.utility) + '.py'
             flags_ = ''
@@ -290,14 +378,52 @@ set -x
             lines.append("""
 export PATH="/usr/lib64/ccache:$PATH"
 """ % vars(self))
+            build_name = 'build_' + srcname
             lines.append(R"""
-python3 -m nuitka  %s %s %s 
-""" % (nflags, flags_, src))
+time python3 -m nuitka  %(nflags)s %(flags_)s %(src)s 2>&1 >%(build_name)s.log
+""" % vars())
             self.fs.folders.append(target_dir)
-            build_name = 'build_' + target_.utility
+            if "outputname" in target_:
+                srcname = target_.outputname
+            lines.append(R"""
+mv  %(target_dir_)s/%(outputname)s   %(target_dir_)s/%(srcname)s 
+""" % vars())
+
+            if "modules" in target_:
+                force_modules = []
+                if 'force_modules' in target_:
+                    force_modules = target_.force_modules
+
+                for it in target_.modules + force_modules:
+                    mdir_ = None
+                    try:
+                        mdir_ = dir4module(it) 
+                        mdir__ = os.path.relpath(mdir_)
+                        if len(mdir__)<len(mdir_):
+                            mdir_ = mdir__
+                    except:
+                        pass                
+
+                    try:
+                        mdir_ = module2build[it].folder 
+                    except:
+                        pass                
+
+                    if mdir_:        
+                        lines.append(R"""
+rsync -rav --exclude=*.py --exclude=*.pyc --exclude=__pycache__ --prune-empty-dirs %(mdir_)s %(target_dir_)s/                
+""" % vars())
+
+                force_modules = []
+                for it in target_.modules:
+                    lines.append(R"""
+rsync -av --include=*.so --include=*.bin --exclude=*  %(tmpdir_)s/modules/%(it)s/ %(target_dir_)s/.                
+rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.                
+""" % vars())
             self.lines2sh(build_name, lines, None)
             bfiles.append(build_name)
 
+
         lines = []
         for b_ in bfiles:
             lines.append("./" + b_ + '.sh')
@@ -392,6 +518,7 @@ python3 -m nuitka  %s %s %s
         '''
         Генерируем список RPM-зависимостей для заданного списка пакетов.
         '''
+        pl_ = self.packages2list(package_list)
         repoch = re.compile("\d+\:")
         def remove_epoch(package):
             package_ = repoch.sub('', package)    
@@ -417,20 +544,20 @@ python3 -m nuitka  %s %s %s
             res = ''
             for try_ in range(3):
                 try:
-                    res = subprocess.check_output(['repoquery'] + options_  + package_list,  universal_newlines=True)
+                    res = subprocess.check_output(['repoquery', '-y'] + options_  + pl_,  universal_newlines=True)
                     break
                 except subprocess.CalledProcessError:
                     #  died with <Signals.SIGSEGV: 11>.
                     time.sleep(2)                    
             # res = subprocess.check_output(['repoquery'] + options_  + ['--output', 'dot-tree'] + package_list,  universal_newlines=True)
             with open(os.path.join(self.start_dir, 'deps.txt'), 'w', encoding='utf-8') as lf:
-                lf.write('\n -'.join(package_list))
+                lf.write('\n -'.join(pl_))
                 lf.write('\n----------------\n')
                 lf.write(res)
     
-        output  = subprocess.check_output(['repoquery'] + options_ + package_list,  universal_newlines=True).splitlines()
+        output  = subprocess.check_output(['repoquery'] + options_ + pl_,  universal_newlines=True).splitlines()
         output = [remove_epoch(x) for x in output if self.ps.is_package_needed(x)]
-        packages_ = output + package_list
+        packages_ = output + pl_
         with open(os.path.join(self.start_dir, 'selected-packages.txt'), 'w', encoding='utf-8') as lf:
             lf.write('\n- '.join(packages_))
 
@@ -479,16 +606,24 @@ python3 -m nuitka  %s %s %s
         # So we need a separate step in which all packages are added together.
     
         for package_ in packages:
-            files = subprocess.check_output(['repoquery',
-                                         '--installed',
-                                         '--cacheonly',
-                                         '--list' ] + [package_], universal_newlines=True).splitlines()
+            for try_ in range(3):
+                try:
+                    files = subprocess.check_output(['repoquery',
+                                                '-y',
+                                                '--installed',
+                                                '--cacheonly',
+                                                '--list' ] + [package_], universal_newlines=True).splitlines()
+                    break                                
+                except:
+                    pass                                            
+
             for file in files:
                 if 'i686' in file:
                     assert(True)
             
         
         candidates = subprocess.check_output(['repoquery',
+                                     '-y',
                                      '--installed',
                                      '--cacheonly',
                                      '--list' ] + packages, universal_newlines=True).splitlines()
@@ -519,7 +654,8 @@ python3 -m nuitka  %s %s %s
             pass
         except Exception as ex_:
             print("Troubles on adding", to_ , "<-", what)
-            raise ex_
+            pass
+            #raise ex_
             pass
 
 
@@ -537,7 +673,7 @@ python3 -m nuitka  %s %s %s
             return
         
         # if m.mime_type not in ['application/x-sharedlib', 'application/x-executable']
-        if not 'application' in m:
+        if not 'ELF' in m:
             return
     
         pyname = os.path.basename(binpath)
@@ -640,14 +776,27 @@ fi
 
         root_dir = self.root_dir
         args = self.args
-        for td_ in self.pp.terra:
+
+        if self.args.debug:
+            pl_ = self.get_wheel_list_to_install()
+            # pls_ = " ".join(pl_)
+            for pls_ in pl_:
+                scmd = '%(root_dir)s/ebin/python3 -m pip install  %(pls_)s ' % vars()
+                print(scmd)
+                os.system(scmd)
+
+        nodes_ = self.pp.terra
+        if self.args.debug:
+            nodes_ += self.pp.build
+        for td_ in nodes_:
             git_url, git_branch, path_to_dir, setup_path = self.explode_pp_node(td_)
     
             os.chdir(setup_path)
             make_setup_if_not_exists()
-            # release_mod = ''
+            release_mod = ''
             # if self.args.release:
-            release_mod = ' --exclude-source-files '
+            if not self.args.debug:
+                release_mod = ' --exclude-source-files '
             scmd = "%(root_dir)s/ebin/python3 setup.py install --single-version-externally-managed  %(release_mod)s --root /   " % vars()
             print(scmd)
             os.system(scmd)
@@ -726,14 +875,12 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
         # os.chdir(self.output_dir)
         pass
 
-    def install_wheels(self):
+    def get_wheel_list_to_install(self):
         os.chdir(self.curdir)
 
-        lines = []
-
         in_bin = os.path.relpath(self.in_bin, start=self.curdir)
 
-        our_whl_path = os.path.join(self.in_bin, "ourwheel")
+        our_whl_path = os.path.join(in_bin, "ourwheel")
         our_wheels = []
         our_wheels_set = set()
         if os.path.exists(our_whl_path):
@@ -746,8 +893,24 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
         if os.path.exists(ext_whl_path):
             ext_wheels = [os.path.join(in_bin, "extwheel", whl) for whl in os.listdir(os.path.join(self.in_bin, "extwheel")) if whl.endswith('.whl') and parse_wheel_filename(whl).project not in our_wheels_set]
             ext_src = [os.path.join(in_bin, "extwheel", whl) for whl in os.listdir(os.path.join(self.in_bin, "extwheel")) if whl.endswith('tar.gz') or whl.endswith('tar.bz2')]
-        scmd = 'sudo python3 -m pip install  %s ' % (" ".join(ext_wheels + our_wheels + ext_src))
+
+        return ext_wheels + our_wheels + ext_src
+
+
+    def install_wheels(self):
+        os.chdir(self.curdir)
+
+        lines = []
+
+        pl_ = self.get_wheel_list_to_install()
+        #--use-feature=2020-resolver
+        scmd = 'sudo python3 -m pip install --force-reinstall  %s ' % (" ".join(pl_))
         lines.append(scmd)
+
+        for p_ in pl_:
+            scmd = 'sudo python3 -m pip install --force-reinstall  %s ' % p_
+            lines.append(scmd)
+
         self.lines2sh("15-install-wheels", lines, "install-wheels")
         pass    
 
@@ -762,14 +925,21 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
         root_dir = self.root_dir
         args = self.args
 
+        bin_dir = os.path.relpath(self.in_bin, start=self.curdir)
+
         lines = []
+        lines.append('''
+rm -f %s/extwheel/*        
+''' % bin_dir)
 
         for td_ in self.pp.pip:
-            scmd = "python3 -m pip download %s --dest %s/extwheel " % (td_, os.path.relpath(self.in_bin, start=self.curdir))
+            scmd = "python3 -m pip download %s --dest %s/extwheel " % (td_, bin_dir)
             lines.append(scmd)                
 
         for td_, local_ in [ (x, True) for x in self.pp.build ] + [(x, False) for x in self.pp.terra]:
             git_url, git_branch, path_to_dir, setup_path = self.explode_pp_node(td_)
+            scmd = 'echo "\n\n** Downloading external wheels for %s" **\n' % path_to_dir
+            lines.append(scmd)                
             if os.path.exists(setup_path):
                 os.chdir(setup_path)
                 scmd = "python3 -m pip download %s --dest %s/extwheel " % (
@@ -786,25 +956,26 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
         parentdir, curname = os.path.split(self.curdir)
         disabled_suffix = curname + '.tar.bz2'
 
+        banned_ext = ['.old', '.iso', disabled_suffix]
+        banned_start = ['tmp']
+        banned_mid = ['/out/', '/wtf/', '/.vagrant/', '/.git/']
+
         def filter_(tarinfo):
-            if tarinfo.name.endswith('.old'):
-                print(tarinfo.name)
-                return None
+            for s in banned_ext:
+                if tarinfo.name.endswith(s):
+                    print(tarinfo.name)
+                    return None
 
-            if tarinfo.name.endswith('.iso'):
-                print(tarinfo.name)
-                return None
+            for s in banned_start:
+                if tarinfo.name.startswith(s):
+                    print(tarinfo.name)
+                    return None
 
-            if tarinfo.name.startswith('tmp'):
-                print(tarinfo.name)
-                return None
+            for s in banned_mid:
+                if s in tarinfo.name:
+                    print(tarinfo.name)
+                    return None
 
-            if tarinfo.name.endswith(disabled_suffix):
-                print(tarinfo.name)
-                return None
-
-            if tarinfo.name in ['.vagrant', '.git']:
-                return None
             return tarinfo          
 
 
@@ -904,10 +1075,20 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
 terrarium_assembler --stage-pack=./out "%(specfile_)s" 
             ''' % vars()])
 
+        self.lines2sh("91-pack-debug", [
+            '''
+terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" 
+            ''' % vars()])
+
         root_dir = 'out'
         if self.args.stage_pack:
             root_dir = self.root_dir = expandpath(args.stage_pack)
-            file_list = self.generate_file_list(self.dependencies(self.ps.terra))
+            packages_to_deploy = self.ps.terra
+            if self.args.debug:
+                packages_to_deploy += self.ps.terra + self.ps.build
+
+
+            file_list = self.generate_file_list(self.dependencies(packages_to_deploy))
 
             os.system('echo 2 > /proc/sys/vm/drop_caches ')
             if os.path.exists(root_dir + ".old"):
@@ -965,7 +1146,8 @@ terrarium_assembler --stage-pack=./out "%(specfile_)s"
                         except Exception as ex_:
                             print("Cannot detect Magic for ", f)    
                             raise ex_
-                        if m.startswith('application/x-sharedlib') or m.startswith('application/x-pie-executable'):
+                        if m.startswith('ELF') and 'shared' in m:
+                        # startswith('application/x-sharedlib') or m.startswith('application/x-pie-executable'):
                             self.fix_sharedlib(f, libfile)
                         else:
                             # in case this is a directory that is listed, we don't want to include everything that is in that directory
